/*
 * Copyright (c) 2013 Giam Teck Choon <choon@choon.net> or CHOON.NET.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <resolv.h>
#include <errno.h>
#include <libowfat/buffer.h>
#include <libowfat/io.h>
#include <libowfat/stralloc.h>
#include <libowfat/byte.h>
#include <libowfat/ip4.h>
#include <libowfat/ip6.h>
#include <libowfat/dns.h>

#include <msg/msg.h>
#include "mailfront.h"

#define eprintf(...) fprintf(stderr, __VA_ARGS__)
#define elogerr(N, M, ...) { if(N){ if(M){ eprintf_logname(); } fprintf(stderr, __VA_ARGS__); } }

#ifndef MAX_DNSIP_RECORDS
#define MAX_DNSIP_RECORDS 10
#endif

static response resp;

static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-helo";
static unsigned long thispid;
static int is_ipv6 = 0;
static const char *local_host;
static const char *local_ip;
static const char *remote_host;
static const char *remote_ip;
static const char *relayclient;
static const char *rejectcheck;
static const char *defercheck;
static const char *skipcheck;
static const char *logcheck;
static const char *addheadercheck;
static const char *checkremotehost;
static const char *checkremotehostmatch;
static const char *checkremoteip;
static const char *checkremoteipmatch;
static int outofmemory = 0;
static str estr;
static str hmsg;

void eprintf_logname()
{
  if (!thispid) {
    thispid = (unsigned long)getpid();
  }
  eprintf("%s-%s[%lu]: ", mainprog, thisprog, thispid);
}

struct HostDNSIP {
  char *hostname;
  int size;
  int ptrcheck;
  int ipv4count;
  int ipv6count;
  int ipcount;
  char *ip[MAX_DNSIP_RECORDS];
  int ptrcount;
  char *ptr[MAX_DNSIP_RECORDS];
};

void init_HostDNSIP(struct HostDNSIP *d)
{
  d->hostname = NULL;
  d->size = 0;
  d->ptrcheck = 0;
  d->ipv4count = 0;
  d->ipv6count = 0;
  d->ipcount = 0;
  d->ptrcount = 0;
}

void free_HostDNSIP(struct HostDNSIP *d)
{
  int x = 0;
  if (d->ipcount > 0) {
    for (x = 0; x < d->ipcount; x++) {
      free(d->ip[x]);
    }
    d->ipcount = 0;
  }
  if (d->ptrcount > 0) {
    for (x = 0; x < d->ptrcount; x++) {
      free(d->ptr[x]);
    }
    d->ptrcount = 0;
  }
  init_HostDNSIP(d);
}

/* return NULL if out of memory */
char *ip_ptr_lookup(char *ip)
{
  struct hostent *hent;
  struct in_addr addr;

  char *hostname = (char *)malloc(256);
  if (hostname == NULL) {
    return NULL;
  }
  sprintf(hostname, "%s", ip);

  if(!inet_aton(ip, &addr))
    return hostname;

  if(strchr(ip, ':')) {
    if((hent = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET6))) {
      strcpy(hostname, hent->h_name);
    }
  } else {
    if((hent = gethostbyaddr((char *)&(addr.s_addr), sizeof(addr.s_addr), AF_INET))) {
      strcpy(hostname, hent->h_name);
    }
  }
  return hostname;
}

/* return 1 if is valid IPv4 or IPv6 otherwise return 0 */
int is_valid_ip(const char *ipstr)
{
  if (strchr(ipstr, ':')) {
    char ip[16];
    if (scan_ip6(ipstr, ip) == strlen(ipstr)) {
      return 1;
    }
  } else {
    char ip[4];
    if (scan_ip4(ipstr, ip) == strlen(ipstr)) {
      return 1;
    }
  }
  return 0;
}

/* return 1 if within the match class otherwise return 0
 * For IPv4, we lookup to see whether it is the same class C
 * For IPv6, we lookup for the first 12 chars whether it is the same
 */
int match_ipclass(const char *ipstr1, const char *ipstr2)
{
  if (strchr(ipstr1, ':')) {
    if (strchr(ipstr2, ':')) {
      char ip1[16];
      char ip2[16];
      if (scan_ip6(ipstr1, ip1) != strlen(ipstr1)) {
        return 0;
      }
      if (scan_ip6(ipstr2, ip2) != strlen(ipstr2)) {
        return 0;
      }
      if (byte_equal(ip1, 12, ip2)) {
        return 1;
      }
    }
    return 0;
  } else {
    if (strchr(ipstr2, ':')) {
      return 0;
    }
    char ip1[4];
    char ip2[4];
    if (scan_ip4(ipstr1, ip1) != strlen(ipstr1)) {
      return 0;
    }
    if (scan_ip4(ipstr2, ip2) != strlen(ipstr2)) {
      return 0;
    }
    if (byte_equal(ip1, 3, ip2)) {
      return 1;
    }
  }
  return 0;
}

void str_tolower(const char *a, char *b) {
  int i, len = strlen(a);
  for(i = 0; i < len; i++) {
    b[i] = (char)tolower(a[i]);
  }
  b[len] = '\0';
}

/*
 * return 1 if hostname DNS lookup has A/AAAA record
 * otherwise return 0 if failed or out of memory which will increment
 * static int outofmemory
 * code reference: http://www.logix.cz/michal/devel/various/getaddrinfo.c.xp
 */
int check_hostname(const char *hostname, struct HostDNSIP *s, int limit)
{
  struct addrinfo hints, *res, *result;
  int ret = 0;
  int found = 0;
  int i = 0;
  int c = 0;
  if (hostname) {
    if (s->hostname) {
      elogerr(logcheck, 1, "(%s:%d) s->hostname already set to %s\n", __FILE__, __LINE__, s->hostname);
      return 1;
    }
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    ret = getaddrinfo(hostname, NULL, &hints, &result);
    if (ret) {
      elogerr(logcheck, 1, "(%s:%d) hostname=%s errno=%d errmsg=%s\n", __FILE__, __LINE__, hostname, ret, gai_strerror(ret));
      return ret;
    } else {
      for (c = 0, res = result; res != NULL && c < limit; res = res->ai_next, c++) {
        char ipstr[INET6_ADDRSTRLEN];
        int x = 0;
        int saved = 1;
        void * addr;
        inet_ntop(res->ai_family, res->ai_addr->sa_data, ipstr, sizeof(ipstr));
        switch(res->ai_family) {
          case AF_INET:
            addr = &((struct sockaddr_in *) res->ai_addr)->sin_addr;
            break;
          case AF_INET6:
            addr = &((struct sockaddr_in6 *) res->ai_addr)->sin6_addr;
            break;
        }
        inet_ntop(res->ai_family, addr, ipstr, sizeof(ipstr));
        if (strcmp(ipstr, hostname)) {
          s->hostname = (char *)hostname;
          found = 1;
        }
        if (i > 0) {
          for (x = 0; x < i; x++) {
            if (strcmp(s->ip[x],
              ipstr) == 0) {
              saved = 0;
            }
          }
        }
        if (saved > 0) {
          s->ip[i] = strdup(ipstr);
          if (s->ip[i] == NULL) {
            elogerr(1, 1, "(%s:%d) Out of memory!\n", __FILE__, __LINE__);
            outofmemory++;
            return 0;
          }
          s->ipcount += 1;
          s->ptr[i] = ip_ptr_lookup(ipstr);
          if (s->ptr[i] == NULL) {
            elogerr(1, 1, "(%s:%d) Out of memory!\n", __FILE__, __LINE__);
            outofmemory++;
            return 0;
          }
          s->ptrcount += 1;
          /* If ip and ptr are the same means PTR lookup failed
           * If hostname and ptr are the same means ptr resolves back to the same hostname
           */
          if (strcmp(s->ip[i], s->ptr[i])) {
            if (strcmp(s->hostname, s->ptr[i]) == 0) {
              s->ptrcheck++;
            }
          }
          /* See ip is ipv6 by looking for character ':' */
          if (strchr(s->ip[i], ':')) {
            s->ipv6count++;
          } else {
            s->ipv4count++;
          }
          i++;
        }
      }
      s->size = i;
      freeaddrinfo(result);
    }
  }
  if (found) {
    return found;
  }
  return 0;
}

static const response *init(void)
{
  local_ip = getprotoenv("LOCALIP");
  remote_ip = getprotoenv("REMOTEIP");
  local_host = getprotoenv("LOCALHOST");
  remote_host = getprotoenv("REMOTEHOST");
  if (remote_ip && *remote_ip) {
    if (strchr(remote_ip, ':'))
      is_ipv6 = 1;
  }
  relayclient = session_getenv("RELAYCLIENT");
  rejectcheck = session_getenv("HELO_CHECK_REJECT");
  defercheck = session_getenv("HELO_CHECK_DEFER");
  skipcheck = session_getenv("HELO_CHECK_SKIP");
  checkremotehost = session_getenv("HELO_CHECK_REMOTEHOST");
  checkremotehostmatch = session_getenv("HELO_CHECK_REMOTEHOST_MATCH");
  checkremoteip = session_getenv("HELO_CHECK_REMOTEIP");
  checkremoteipmatch = session_getenv("HELO_CHECK_REMOTEIP_MATCH");
  if (session_getenv("SENDER_HELO_HOSTNAME") != 0) {
    unsetenv("SENDER_HELO_HOSTNAME");
  }
  logcheck = session_getenv("HELO_CHECK_VERBOSE");
  addheadercheck = session_getenv("HELO_CHECK_ADD_HEADER");
  thispid = (unsigned long)getpid();
  return 0;
}

static const response *helo(str *hostname, str *capabilities)
{
  if (session_getenv("SENDER_HELO_HOSTNAME") == 0) {
    char t[strlen(hostname->s) + 1];
    str_tolower(hostname->s, t);
    session_setenv("SENDER_HELO_HOSTNAME", t, 0);
  }
  return 0;
  (void)capabilities;
}

int check_senderhelo_char_start(const char *s, int i, char c)
{
  char d = '\'';
  if (c == '\'') {
    d = '"';
  }
  /* first character */
  if (*s == c) {
    if (i > 0) {
      char *tmp = (char *)malloc(100);
      if (tmp == NULL) {
        outofmemory++;
        return 0;
      }
      sprintf(tmp, "Sorry, your HELO name starts with disallowed character %c%c%c", d, c, d);
      str_init(&estr);
      str_cats(&estr, tmp);
      resp.message = estr.s;
      free(tmp);
      return 2;
    } else {
      elogerr(logcheck, 1, "%c%c%c found in first character of HELO name %s\n", d, c, d, s);
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: ");
      str_catc(&hmsg, d);
      str_catc(&hmsg, c);
      str_catc(&hmsg, d);
      str_cats(&hmsg, " found in first character of HELO name ");
      str_cats(&hmsg, s);
      str_cats(&hmsg, "\n");
      return 1;
    }
  }
  return 0;
}

int check_senderhelo_char_end(const char *s, int i, char c)
{
  char d = '\'';
  if (c == '\'') {
    d = '"';
  }
  int len = strlen(s);
  /* last character */
  if (s[len-1] == c) {
    if (i > 0) {
      char *tmp = (char *)malloc(100);
      if (tmp == NULL) {
        outofmemory++;
        return 0;
      }
      sprintf(tmp, "Sorry, your HELO name ends with disallowed character %c%c%c", d, c, d);
      str_init(&estr);
      str_cats(&estr, tmp);
      resp.message = estr.s;
      free(tmp);
      return 2;
    } else {
      elogerr(logcheck, 1, "%c%c%c found in last character of HELO name %s\n", d, c, d, s);
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: ");
      str_catc(&hmsg, d);
      str_catc(&hmsg, c);
      str_catc(&hmsg, d);
      str_cats(&hmsg, " found in last character of HELO name ");
      str_cats(&hmsg, s);
      str_cats(&hmsg, "\n");
      return 1;
    }
  }
  return 0;
}

int check_senderhelo_char(const char *s, int i, char c)
{
  if (strchr(s, c)) {
    char d = '\'';
    if (c == '\'') {
      d = '"';
    }
    if (i > 0) {
      char *tmp = (char *)malloc(100);
      if (tmp == NULL) {
        outofmemory++;
        return 0;
      }
      sprintf(tmp, "Sorry, your HELO name has the disallowed character %c%c%c", d, c, d);
      str_init(&estr);
      str_cats(&estr, tmp);
      resp.message = estr.s;
      free(tmp);
      return 2;
    } else {
      elogerr(logcheck, 1, "%c%c%c character found in HELO name %s\n", d, c, d, s);
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: ");
      str_catc(&hmsg, d);
      str_catc(&hmsg, c);
      str_catc(&hmsg, d);
      str_cats(&hmsg, " character found in HELO name ");
      str_cats(&hmsg, s);
      str_cats(&hmsg, "\n");
      return 1;
    }
  }
  return 0;
}

static const response *sender(str* address, str* params)
{
  if (skipcheck) return 0;
  int action = 0;
  if (rejectcheck || defercheck) {
    if (rejectcheck) {
      action = 553;
      resp.number = action;
    } else {
      action = 451;
      resp.number = action;
    }
  }
  if (!local_ip) {
    local_ip = getprotoenv("LOCALIP");
  }
  if (!local_ip) {
    elogerr(logcheck, 1, "LOCALIP not found thus this plugin skipped\n");
    return 0;
  }
  local_host = getprotoenv("LOCALHOST");
  if (!remote_ip) {
    remote_ip = getprotoenv("REMOTEIP");
  }
  if (!remote_ip) {
    elogerr(logcheck, 1, "REMOTEIP not found thus this plugin skipped\n");
    return 0;
  }
  str_init(&hmsg);
  str_copys(&hmsg, "X-HELO-Check-Result: ");
  /* X-HELO-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4 */
  if (relayclient) {
    elogerr(logcheck, 1, "$RELAYCLIENT is set thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: $RELAYCLIENT is set for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-HELO-Check-Result: SKIPPED: authenticated user from 1.2.3.4 */
  if (session_getnum("authenticated", 0)) {
    elogerr(logcheck, 1, "authenticated user thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: authenticated user from ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-HELO-Check-Result: SKIPPED: LOCALIP is the same as REMOTEIP 1.2.3.4 */
  if (strcmp(local_ip, remote_ip) == 0) {
    elogerr(logcheck, 1, "LOCALIP=REMOTEIP=%s thus skipped\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: LOCALIP is the same as REMOTEIP ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  /* X-HELO-Check-Result: SKIPPED: $RBLSMTPD is set and empty for 1.1.1.1 */
  const char *rblsmtpd = getenv("RBLSMTPD");
  if (rblsmtpd && *rblsmtpd == 0) {
    elogerr(logcheck, 1, "$RBLSMTPD is set and empty thus skipped for REMOTEIP %s\n", remote_ip);
    str_cats(&hmsg, "SKIPPED: $RBLSMTPD is set and empty for ");
    str_cats(&hmsg, remote_ip);
    str_cats(&hmsg, "\n");
    return 0;
  }
  if (!remote_host) {
    remote_host = getprotoenv("REMOTEHOST");
  }
  if ((checkremotehost || checkremotehostmatch) && (!remote_host || *remote_host == 0)) {
    if (checkremotehost && checkremotehostmatch) {
      elogerr(logcheck, 1, "$HELO_CHECK_REMOTEHOST and $HELO_CHECK_REMOTEHOST_MATCH are set for REMOTEIP %s\n", remote_ip);
    } else if (checkremotehost) {
      elogerr(logcheck, 1, "$HELO_CHECK_REMOTEHOST is set for REMOTEIP %s\n", remote_ip);
    } else {
      elogerr(logcheck, 1, "$HELO_CHECK_REMOTEHOST_MATCH is set for REMOTEIP %s\n", remote_ip);
    }
    elogerr(logcheck, 1, "REMOTEHOST is empty thus trying to do PTR DNS lookup from REMOTEIP %s\n", remote_ip);
    char *tmphost = ip_ptr_lookup((char *)remote_ip);
    if (outofmemory > 0 || tmphost == NULL) {
      elogerr(1, 1, "(%s:%d) char *tmphost = ip_ptr_lookup((char *)remote_ip) failed!\n", __FILE__, __LINE__);
      return &resp_oom;
    }
    if (strcmp(tmphost, remote_ip) != 0 && strlen(tmphost) > 0) {
      remote_host = strdup(tmphost);
      if (remote_host == NULL) {
        elogerr(1, 1, "(%s:%d) remote_host=strdup(tmphost) failed!\n", __FILE__, __LINE__);
        free(tmphost);
        return &resp_oom;
      }
      elogerr(logcheck, 1, "PTR DNS loopkup of REMOTEIP=%s success and set REMOTEHOST=%s\n", remote_ip, tmphost);
      free(tmphost);
    } else {
      free(tmphost);
      elogerr(logcheck, 1, "PTR DNS loopkup for REMOTEIP %s failed\n", remote_ip);
      if (action > 0) {
        str_init(&estr);
        str_cats(&estr, "Sorry, your hostname is empty and your ip ");
        str_cats(&estr, remote_ip);
        str_cats(&estr, " doesn't have any PTR record");
        resp.message = estr.s;
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEHOST is empty and PTR DNS lookup for IP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " failed\n");
        return 0;
      }
    }
  }
  const char *senderhelo = session_getenv("SENDER_HELO_HOSTNAME");
  /* If helo_hostname is empty */
  if (senderhelo == 0 || *senderhelo == 0) {
    if (action > 0) {
      resp.message = "Sorry, your HELO name is empty!";
      return &resp;
    } else {
      elogerr(logcheck, 1, "helo hostname is empty\n");
      str_cats(&hmsg, "FAILED\n");
      str_cats(&hmsg, "X-HELO-Check-Summary: HELO name is empty!\n");
      return 0;
    }
  }
  if (senderhelo && *senderhelo != 0) {
    if (checkremotehostmatch) {
      if (strcmp(senderhelo, remote_host) != 0) {
        elogerr(logcheck, 1, "HELO name %s and REMOTEHOST %s doesn't match for REMOTEIP %s\n", senderhelo, remote_host, remote_ip);
        if (action > 0) {
          str_init(&estr);
          str_cats(&estr, "Sorry, your HELO name ");
          str_cats(&estr, senderhelo);
          str_cats(&estr, " and REMOTEHOST ");
          str_cats(&estr, remote_host);
          str_cats(&estr, " doesn't match");
          resp.message = estr.s;
          return &resp;
        } else {
          str_cats(&hmsg, "FAILED\n");
          str_cats(&hmsg, "X-HELO-Check-Summary: HELO name ");
          str_cats(&hmsg, senderhelo);
          str_cats(&hmsg, " REMOTEHOST ");
          str_cats(&hmsg, remote_host);
          str_cats(&hmsg, " doesn't match\n");
          return 0;
        }
      } else {
        elogerr(logcheck, 1, "HELO name %s and REMOTEHOST %s match for REMOTEIP %s\n", senderhelo, remote_host, remote_ip);
      }
    }
    int len = strlen(senderhelo);
    /* 0 continue, 1 to return 0 and 2 to return &resp; */
    int rc = 0;
    /* Don't accept start with a dot */
    rc = check_senderhelo_char_start(senderhelo, action, '.');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    /* Don't accept start with a [ */
    rc = check_senderhelo_char_start(senderhelo, action, '[');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    /* Don't accept start with a dash */
    rc = check_senderhelo_char_start(senderhelo, action, '-');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    if (len > 1) {
      /* Don't accept end with a dot */
      rc = check_senderhelo_char_end(senderhelo, action, '.');
      if (rc == 2) return &resp;
      if (rc == 1) return 0;
      /* Don't accept end with a ] */
      rc = check_senderhelo_char_end(senderhelo, action, ']');
      if (rc == 2) return &resp;
      if (rc == 1) return 0;
      /* Don't accept end with a dash */
      rc = check_senderhelo_char_end(senderhelo, action, '-');
      if (rc == 2) return &resp;
      if (rc == 1) return 0;
    }
    if (strchr(senderhelo, '.') == NULL) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name doesn't contains the character '.'";
        return &resp;
      } else {
        elogerr(logcheck, 1, "'.' character not found in HELO name %s for REMOTEIP %s\n", senderhelo, remote_ip);
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: '.' character not found in HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        return 0;
      }
    }
    rc = check_senderhelo_char(senderhelo, action, '!');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '@');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '#');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '$');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '%');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '^');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '&');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '*');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '(');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ')');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '[');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ']');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '{');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '}');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ':');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ';');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '"');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '\'');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '<');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '>');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '?');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '/');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '|');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '\\');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '+');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '=');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '_');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ' ');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, '\t');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    rc = check_senderhelo_char(senderhelo, action, ',');
    if (rc == 2) return &resp;
    if (rc == 1) return 0;
    if (is_valid_ip(senderhelo)) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name is a valid IP address which is disallowed";
        return &resp;
      } else {
        elogerr(logcheck, 1, "valid IP address found in HELO name %s for REMOTEIP %s\n", senderhelo, remote_ip);
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: Valid IP address found in HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        return 0;
      }
    }
  }
  int matchipcount = 0;
  int matchipclasscount = 0;
  char *matchip = NULL;
  char *senderheloip = NULL;
  struct HostDNSIP hh;
  init_HostDNSIP(&hh);
  elogerr(logcheck, 1, "Start doing DNS Lookup for HELO name %s REMOTEIP %s\n", senderhelo, remote_ip);
  int ret = check_hostname(senderhelo, &hh, MAX_DNSIP_RECORDS);
  if (ret > 0) {
    int x = 0;
    for (x = 0; x < hh.ipcount && hh.ip[x] != NULL; x++) {
      elogerr(logcheck, 1, "ip[%d]=%s ptr[%d]=%s", x, hh.ip[x], x, hh.ptr[x]);
      if (!senderheloip) {
        senderheloip = hh.ip[x];
      }
      /* If it is different then it is a ptr */
      if (strcmp(hh.ip[x], hh.ptr[x]) != 0) {
        elogerr(logcheck, 0, " ptr=YES");
      } else {
        elogerr(logcheck, 0, " ptr=NO");
      }
      if (strcmp(remote_ip, hh.ip[x]) == 0) {
        elogerr(logcheck, 0, " ipmatch=YES");
        matchipcount++;
      }
      else {
        elogerr(logcheck, 0, " ipmatch=NO");
        if (match_ipclass(remote_ip, hh.ip[x])) {
          elogerr(logcheck, 0, " ipclassmatch=YES");
          if (!matchip) {
            matchip = hh.ip[x];
          }
          matchipclasscount++;
        }
      }
      /* If it is the same then it is valid ptr */
      if (strcmp(hh.hostname, hh.ptr[x]) == 0) {
        elogerr(logcheck, 0, " ptrcheck=PASSED\n");
      } else {
        elogerr(logcheck, 0, " ptrcheck=FAILED\n");
      }
    }
    if (checkremoteip || checkremoteipmatch) {
      if (checkremoteip && checkremoteipmatch) {
        elogerr(logcheck, 1, "$HELO_CHECK_REMOTEIP and $HELO_CHECK_REMOTE IP_MATCH are set for REMOTEIP %s\n", remote_ip);
      } else if (checkremoteip) {
        elogerr(logcheck, 1, "$HELO_CHECK_REMOTEIP is set for REMOTEIP %s\n", remote_ip);
      } else {
        elogerr(logcheck, 1, "$HELO_CHECK_REMOTEIP_MATCH is set for REMOTEIP %s\n", remote_ip);
      }
      if (matchipcount > 0) {
        elogerr(logcheck, 1, "REMOTEIP and HELO name %s's IP %s match\n", senderhelo, remote_ip);
        str_cats(&hmsg, "PASSED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " matches ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "'s IP\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      } else if (matchipclasscount > 0 && checkremoteipmatch && matchip) {
        elogerr(logcheck, 1, "REMOTEIP %s and HELO name %s's IP %s are within the same IP class\n", remote_ip, senderhelo, matchip);
        str_cats(&hmsg, "SKIPPED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " and ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "'s IP ");
        str_cats(&hmsg, matchip);
        str_cats(&hmsg, " are within the same IP class\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
      if (checkremoteipmatch) {
        elogerr(logcheck, 1, "REMOTEIP %s and HELO name %s's IP %s are not within the same IP class\n", remote_ip, senderhelo, senderheloip);
        if (action > 0) {
          str_init(&estr);
          str_cats(&estr, "Sorry, your REMOTEIP ");
          str_cats(&estr, remote_ip);
          str_cats(&estr, " doesn't match your HELO name's IP ");
          str_cats(&estr, senderheloip);
          str_cats(&estr, " or within the same IP class");
          resp.message = estr.s;
          return &resp;
        } else {
          str_cats(&hmsg, "FAILED\n");
          str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
          str_cats(&hmsg, remote_ip);
          str_cats(&hmsg, " and ");
          str_cats(&hmsg, senderhelo);
          str_cats(&hmsg, "'s IP ");
          str_cats(&hmsg, senderheloip);
          str_cats(&hmsg, " are not within the same IP class\n");
          return 0;
        }
      }
      elogerr(logcheck, 1, "REMOTEIP %s doesn't match HELO name %s's IP %s\n", remote_ip, senderhelo, senderheloip);
      if (action > 0) {
        str_init(&estr);
        str_cats(&estr, "Sorry, your REMOTEIP ");
        str_cats(&estr, remote_ip);
        str_cats(&estr, " doesn't match your HELO name's IP ");
        str_cats(&estr, senderheloip);
        resp.message = estr.s;
        (void)free_HostDNSIP(&hh);
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: REMOTEIP ");
        str_cats(&hmsg, remote_ip);
        str_cats(&hmsg, " doesn't match ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "'s IP ");
        str_cats(&hmsg, senderheloip);
        str_cats(&hmsg, "\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
    }
    const char *helo_ptrcheck = session_getenv("HELO_CHECK_PTR");
    if (hh.ptrcount == 0 && helo_ptrcheck) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name doesn't has a valid PTR record";
        (void)free_HostDNSIP(&hh);
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: PTR record not found for HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
    }
    const char *helo_ptrcheck_match = session_getenv("HELO_CHECK_PTR_MATCH");
    if (hh.ptrcheck == 0 && helo_ptrcheck_match) {
      if (action > 0) {
        resp.message = "Sorry, your HELO name doesn't has a valid PTR record that resolves back to the same name";
        (void)free_HostDNSIP(&hh);
        return &resp;
      } else {
        str_cats(&hmsg, "FAILED\n");
        str_cats(&hmsg, "X-HELO-Check-Summary: PTR record doesn't resolves back to HELO name ");
        str_cats(&hmsg, senderhelo);
        str_cats(&hmsg, "\n");
        (void)free_HostDNSIP(&hh);
        return 0;
      }
    }
    if (hh.ptrcheck > 0) {
      str_cats(&hmsg, "PASSED\n");
    } else {
      str_cats(&hmsg, "FAILED\n");
    }
    char tt[512];
    str_cats(&hmsg, "X-HELO-Check-Summary: ");
    str_cats(&hmsg, hh.hostname);
    str_cats(&hmsg, " has ");
    sprintf(tt, "%d", hh.ipv4count);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " A ");
    if (hh.ipv4count > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, ", ");
    sprintf(tt, "%d", hh.ipv6count);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " AAAA ");
    if (hh.ipv6count > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, ", ");
    sprintf(tt, "%d", hh.ptrcount);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " PTR ");
    if (hh.ptrcount > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, " and ");
    sprintf(tt, "%d", hh.ptrcheck);
    str_cats(&hmsg, tt);
    str_cats(&hmsg, " PTR ");
    if (hh.ptrcheck > 1) {
      str_cats(&hmsg, "records");
    } else {
      str_cats(&hmsg, "record");
    }
    str_cats(&hmsg, " resolves to the same hostname\n");
    (void)free_HostDNSIP(&hh);
    return 0;
  }
  if (outofmemory > 0 || ret == EAI_MEMORY) {
    eprintf_logname();
    eprintf("(%s:%d) Out of Memory!\n", __FILE__, __LINE__);
    return &resp_oom;
  }
  char *tmpmsg = (char *)malloc(512);
  if (tmpmsg == NULL) {
    elogerr(1, 1, "(%s:%d) char *tmpmsg is out of memory!\n", __FILE__, __LINE__);
    return &resp_oom;
  }
  if (action > 0) {
    if (ret == EAI_NONAME) {
      sprintf(tmpmsg, "Sorry, your HELO name is invalid.");
      if (strlen(senderhelo) > 0) {
        sprintf(tmpmsg, "Sorry, your HELO name %s is invalid.", senderhelo);
      }
      str_init(&estr);
      str_cats(&estr, tmpmsg);
      resp.message = estr.s;
      free(tmpmsg);
      return &resp;
    } else {
      sprintf(tmpmsg, "Sorry, your HELO name DNS lookup returned a temporary failure. Please try again later.");
      if (strlen(senderhelo) > 0) {
        sprintf(tmpmsg, "Sorry, your HELO name %s DNS lookup returned a temporary failure. Please try again later.", senderhelo);
      }
      resp.number = 451;
      str_init(&estr);
      str_cats(&estr, tmpmsg);
      resp.message = estr.s;
      free(tmpmsg);
      return &resp;
    }
  } else {
    str_cats(&hmsg, "FAILED\n");
    elogerr(logcheck, 1, "(%s:%d) HELO name %s DNS lookup error\n", __FILE__, __LINE__, senderhelo);
    str_cats(&hmsg, "X-HELO-Check-Summary: DNS lookup error for HELO name ");
    if (strlen(senderhelo) > 0) {
      str_cats(&hmsg, senderhelo);
    }
    str_cats(&hmsg, "\n");
    free(tmpmsg);
    return 0;
  }
  return 0;
  (void)address;
  (void)params;
}

static const response *data_start(int fd)
{
  if (skipcheck) return 0;
  if (hmsg.len > 0 && addheadercheck) {
    return backend_data_block(hmsg.s, hmsg.len);
  }
  str_init(&hmsg);
  return 0;
  (void)fd;
}

struct plugin plugin = {
  .version = PLUGIN_VERSION,
  .flags = 0,
  .init = init,
  .helo = helo,
  .sender = sender,
  .data_start = data_start,
};
